# A parser for XML, using the derived class as static DTD.
# Author: Sjoerd Mullender.

# sgmlop support added by fredrik@pythonware.com (May 19, 1998)

import re
import string

try:
    import sgmlop
except ImportError:
    sgmlop = None

# standard entity defs

ENTITYDEFS = {
    'lt': '<',
    'gt': '>',
    'amp': '&',
    'quot': '"',
    'apos': '\''
    }

# XML parser base class -- find tags and call handler functions.
# Usage: p = XMLParser(); p.feed(data); ...; p.close().
# The dtd is defined by deriving a class which defines methods with
# special names to handle tags: start_foo and end_foo to handle <foo>
# and </foo>, respectively.  The data between tags is passed to the
# parser by calling self.handle_data() with some data as argument (the
# data may be split up in arbutrary chunks).  Entity references are
# passed by calling self.handle_entityref() with the entity reference
# as argument.

# --------------------------------------------------------------------
# original re-based XML parser

_S = '[ \t\r\n]+'
_opS = '[ \t\r\n]*'
_Name = '[a-zA-Z_:][-a-zA-Z0-9._:]*'
interesting = re.compile('[&<]')
incomplete = re.compile('&(' + _Name + '|#[0-9]*|#x[0-9a-fA-F]*)?|'
			   '<([a-zA-Z_:][^<>]*|'
			      '/([a-zA-Z_:][^<>]*)?|'
			      '![^<>]*|'
			      '\?[^<>]*)?')

ref = re.compile('&(' + _Name + '|#[0-9]+|#x[0-9a-fA-F]+);?')
entityref = re.compile('&(?P<name>' + _Name + ')[^-a-zA-Z0-9._:]')
charref = re.compile('&#(?P<char>[0-9]+[^0-9]|x[0-9a-fA-F]+[^0-9a-fA-F])')
space = re.compile(_S)
newline = re.compile('\n')

starttagopen = re.compile('<' + _Name)
endtagopen = re.compile('</')
starttagend = re.compile(_opS + '(?P<slash>/?)>')
endbracket = re.compile('>')
tagfind = re.compile(_Name)
cdataopen = re.compile('<!\[CDATA\[')
cdataclose = re.compile('\]\]>')
special = re.compile('<!(?P<special>[^<>]*)>')
procopen = re.compile('<\?(?P<proc>' + _Name + ')' + _S)
procclose = re.compile('\?>')
commentopen = re.compile('<!--')
commentclose = re.compile('-->')
doubledash = re.compile('--')
attrfind = re.compile(
    _opS + '(?P<name>' + _Name + ')'
    '(' + _opS + '=' + _opS +
    '(?P<value>\'[^\']*\'|"[^"]*"|[-a-zA-Z0-9.:+*%?!()_#=~]+))')

class SlowXMLParser:

    # Interface -- initialize and reset this instance
    def __init__(self, verbose=0):
	self.verbose = verbose
	self.reset()

    # Interface -- reset this instance.  Loses all unprocessed data
    def reset(self):
	self.rawdata = ''
	self.stack = []
	self.lasttag = '???'
	self.nomoretags = 0
	self.literal = 0
	self.lineno = 1

    # For derived classes only -- enter literal mode (CDATA) till EOF
    def setnomoretags(self):
	self.nomoretags = self.literal = 1

    # For derived classes only -- enter literal mode (CDATA)
    def setliteral(self, *args):
	self.literal = 1

    # Interface -- feed some data to the parser.  Call this as
    # often as you want, with as little or as much text as you
    # want (may include '\n').  (This just saves the text, all the
    # processing is done by goahead().)
    def feed(self, data):
	self.rawdata = self.rawdata + data
	self.goahead(0)

    # Interface -- handle the remaining data
    def close(self):
	self.goahead(1)

    # Interface -- translate references
    def translate_references(self, data):
	newdata = []
	i = 0
	while 1:
	    res = ref.search(data, i)
	    if res is None:
		newdata.append(data[i:])
		return string.join(newdata, '')
	    if data[res.end(0) - 1] != ';':
		self.syntax_error(self.lineno,
				  '; missing after entity/char reference')
	    newdata.append(data[i:res.start(0)])
	    str = res.group(1)
	    if str[0] == '#':
		if str[1] == 'x':
		    newdata.append(chr(string.atoi(str[2:], 16)))
		else:
		    newdata.append(chr(string.atoi(str[1:])))
	    else:
		try:
		    newdata.append(self.entitydefs[str])
		except KeyError:
		    # can't do it, so keep the entity ref in
		    newdata.append('&' + str + ';')
	    i = res.end(0)

    # Internal -- handle data as far as reasonable.  May leave state
    # and data to be processed by a subsequent call.  If 'end' is
    # true, force handling all data as if followed by EOF marker.
    def goahead(self, end):
	rawdata = self.rawdata
	i = 0
	n = len(rawdata)
	while i < n:
	    if self.nomoretags:
		data = rawdata[i:n]
		self.handle_data(data)
		self.lineno = self.lineno + string.count(data, '\n')
		i = n
		break
	    res = interesting.search(rawdata, i)
	    if res:
		    j = res.start(0)
	    else:
		    j = n
	    if i < j:
		data = rawdata[i:j]
		self.handle_data(data)
		self.lineno = self.lineno + string.count(data, '\n')
	    i = j
	    if i == n: break
	    if rawdata[i] == '<':
		if starttagopen.match(rawdata, i):
		    if self.literal:
			data = rawdata[i]
			self.handle_data(data)
			self.lineno = self.lineno + string.count(data, '\n')
			i = i+1
			continue
		    k = self.parse_starttag(i)
		    if k < 0: break
		    self.lineno = self.lineno + string.count(rawdata[i:k], '\n')
		    i = k
		    continue
		if endtagopen.match(rawdata, i):
		    k = self.parse_endtag(i)
		    if k < 0: break
		    self.lineno = self.lineno + string.count(rawdata[i:k], '\n')
		    i =  k
		    self.literal = 0
		    continue
		if commentopen.match(rawdata, i):
		    if self.literal:
			data = rawdata[i]
			self.handle_data(data)
			self.lineno = self.lineno + string.count(data, '\n')
			i = i+1
			continue
		    k = self.parse_comment(i)
		    if k < 0: break
		    self.lineno = self.lineno + string.count(rawdata[i:k], '\n')
		    i = k
		    continue
		if cdataopen.match(rawdata, i):
		    k = self.parse_cdata(i)
		    if k < 0: break
		    self.lineno = self.lineno + string.count(rawdata[i:i], '\n')
		    i = k
		    continue
		res = procopen.match(rawdata, i)
		if res:
		    k = self.parse_proc(i, res)
		    if k < 0: break
		    self.lineno = self.lineno + string.count(rawdata[i:k], '\n')
		    i = k
		    continue
		res = special.match(rawdata, i)
		if res:
		    if self.literal:
			data = rawdata[i]
			self.handle_data(data)
			self.lineno = self.lineno + string.count(data, '\n')
			i = i+1
			continue
		    self.handle_special(res.group('special'))
		    self.lineno = self.lineno + string.count(res.group(0), '\n')
		    i = res.end(0)
		    continue
	    elif rawdata[i] == '&':
		res = charref.match(rawdata, i)
		if res is not None:
		    i = res.end(0)
		    if rawdata[i-1] != ';':
			self.syntax_error(self.lineno, '; missing in charref')
			i = i-1
		    self.handle_charref(res.group('char')[:-1])
		    self.lineno = self.lineno + string.count(res.group(0), '\n')
		    continue
		res = entityref.match(rawdata, i)
		if res is not None:
		    i = res.end(0)
		    if rawdata[i-1] != ';':
			self.syntax_error(self.lineno, '; missing in entityref')
			i = i-1
		    self.handle_entityref(res.group('name'))
		    self.lineno = self.lineno + string.count(res.group(0), '\n')
		    continue
	    else:
		raise RuntimeError, 'neither < nor & ??'
	    # We get here only if incomplete matches but
	    # nothing else
	    res = incomplete.match(rawdata, i)
	    if not res:
		data = rawdata[i]
		self.handle_data(data)
		self.lineno = self.lineno + string.count(data, '\n')
		i = i+1
		continue
	    j = res.end(0)
	    if j == n:
		break # Really incomplete
	    self.syntax_error(self.lineno, 'bogus < or &')
	    data = res.group(0)
	    self.handle_data(data)
	    self.lineno = self.lineno + string.count(data, '\n')
	    i = j
	# end while
	if end and i < n:
	    data = rawdata[i:n]
	    self.handle_data(data)
	    self.lineno = self.lineno + string.count(data, '\n')
	    i = n
	self.rawdata = rawdata[i:]
	# XXX if end: check for empty stack

    # Internal -- parse comment, return length or -1 if not terminated
    def parse_comment(self, i):
	rawdata = self.rawdata
	if rawdata[i:i+4] <> '<!--':
	    raise RuntimeError, 'unexpected call to handle_comment'
	res = commentclose.search(rawdata, i+4)
	if not res:
	    return -1
	# doubledash search will succeed because it's a subset of commentclose
	if doubledash.search(rawdata, i+4).start(0) < res.start(0):
	    self.syntax_error(self.lineno, "`--' inside comment")
	self.handle_comment(rawdata[i+4: res.start(0)])
	return res.end(0)

    # Internal -- handle CDATA tag, return lenth or -1 if not terminated
    def parse_cdata(self, i):
	rawdata = self.rawdata
	if rawdata[i:i+9] <> '<![CDATA[':
	    raise RuntimeError, 'unexpected call to handle_cdata'
	res = cdataclose.search(rawdata, i+9)
	if not res:
	    return -1
	self.handle_cdata(rawdata[i+9:res.start(0)])
	return res.end(0)

    def parse_proc(self, i, res):
	rawdata = self.rawdata
	if not res:
	    raise RuntimeError, 'unexpected call to parse_proc'
	name = res.group('proc')
	res = procclose.search(rawdata, res.end(0))
	if not res:
	    return -1
	self.handle_proc(name, rawdata[res.pos:res.start(0)])
	return res.end(0)

    # Internal -- handle starttag, return length or -1 if not terminated
    def parse_starttag(self, i):
	rawdata = self.rawdata
	# i points to start of tag
	end = endbracket.search(rawdata, i+1)
	if not end:
	    return -1
	j = end.start(0)
	# Now parse the data between i+1 and j into a tag and attrs
	attrdict = {}
	res = tagfind.match(rawdata, i+1)
	if not res:
	    raise RuntimeError, 'unexpected call to parse_starttag'
	k = res.end(0)
	tag = res.group(0)
	if hasattr(self, tag + '_attributes'):
	    attrlist = getattr(self, tag + '_attributes')
	else:
	    attrlist = None
	self.lasttag = tag
	while k < j:
	    res = attrfind.match(rawdata, k)
	    if not res: break
	    attrname, attrvalue = res.group('name', 'value')
	    if attrvalue is None:
		self.syntax_error(self.lineno, 'no attribute value specified')
		attrvalue = attrname
	    elif attrvalue[:1] == "'" == attrvalue[-1:] or \
		 attrvalue[:1] == '"' == attrvalue[-1:]:
		attrvalue = attrvalue[1:-1]
	    else:
		self.syntax_error(self.lineno, 'attribute value not quoted')
	    if attrlist is not None and attrname not in attrlist:
		self.syntax_error(self.lineno,
				  'unknown attribute %s of element %s' %
				  (attrname, tag))
	    if attrdict.has_key(attrname):
		self.syntax_error(self.lineno, 'attribute specified twice')
	    attrdict[attrname] = self.translate_references(attrvalue)
	    k = res.end(0)
	res = starttagend.match(rawdata, k)
	if not res:
	    self.syntax_error(self.lineno, 'garbage in start tag')
	self.finish_starttag(tag, attrdict)
	if res and res.group('slash') == '/':
	    self.finish_endtag(tag)
	return end.end(0)

    # Internal -- parse endtag
    def parse_endtag(self, i):
	rawdata = self.rawdata
	end = endbracket.search(rawdata, i+1)
	if not end:
	    return -1
	res = tagfind.match(rawdata, i+2)
	if not res:
	    self.syntax_error(self.lineno, 'no name specified in end tag')
	    tag = ''
	    k = i+2
	else:
	    tag = res.group(0)
	    k = res.end(0)
	if k != end.start(0):
	    # check that there is only white space at end of tag
	    res = space.match(rawdata, k)
	    if res is None or res.end(0) != end.start(0):
		self.syntax_error(self.lineno, 'garbage in end tag')
	self.finish_endtag(tag)
	return end.end(0)

    # Internal -- finish processing of start tag
    # Return -1 for unknown tag, 1 for balanced tag
    def finish_starttag(self, tag, attrs):
	self.stack.append(tag)
	try:
	    method = getattr(self, 'start_' + tag)
	except AttributeError:
	    self.unknown_starttag(tag, attrs)
	    return -1
	else:
	    self.handle_starttag(tag, method, attrs)
	    return 1

    # Internal -- finish processing of end tag
    def finish_endtag(self, tag):
	if not tag:
	    found = len(self.stack) - 1
	    if found < 0:
		self.unknown_endtag(tag)
		return
	else:
	    if tag not in self.stack:
		try:
		    method = getattr(self, 'end_' + tag)
		except AttributeError:
		    self.unknown_endtag(tag)
		return
	    found = len(self.stack)
	    for i in range(found):
		if self.stack[i] == tag: found = i
	while len(self.stack) > found:
	    tag = self.stack[-1]
	    try:
		method = getattr(self, 'end_' + tag)
	    except AttributeError:
		method = None
	    if method:
		self.handle_endtag(tag, method)
	    else:
		self.unknown_endtag(tag)
	    del self.stack[-1]

    # Overridable -- handle start tag
    def handle_starttag(self, tag, method, attrs):
	method(attrs)

    # Overridable -- handle end tag
    def handle_endtag(self, tag, method):
	method()

    # Example -- handle character reference, no need to override
    def handle_charref(self, name):
	try:
	    if name[0] == 'x':
		n = string.atoi(name[1:], 16)
	    else:
		n = string.atoi(name)
	except string.atoi_error:
	    self.unknown_charref(name)
	    return
	if not 0 <= n <= 255:
	    self.unknown_charref(name)
	    return
	self.handle_data(chr(n))

    # Definition of entities -- derived classes may override
    entitydefs = ENTITYDEFS

    # Example -- handle entity reference, no need to override
    def handle_entityref(self, name):
	table = self.entitydefs
	if table.has_key(name):
	    self.handle_data(table[name])
	else:
	    self.unknown_entityref(name)
	    return

    # Example -- handle data, should be overridden
    def handle_data(self, data):
	pass

    # Example -- handle cdata, could be overridden
    def handle_cdata(self, data):
	pass

    # Example -- handle comment, could be overridden
    def handle_comment(self, data):
	pass

    # Example -- handle processing instructions, could be overridden
    def handle_proc(self, name, data):
	pass

    # Example -- handle special instructions, could be overridden
    def handle_special(self, data):
	pass

    # Example -- handle relatively harmless syntax errors, could be overridden
    def syntax_error(self, lineno, message):
	raise RuntimeError, 'Syntax error at line %d: %s' % (lineno, message)

    # To be overridden -- handlers for unknown objects
    def unknown_starttag(self, tag, attrs): pass
    def unknown_endtag(self, tag): pass
    def unknown_charref(self, ref): pass
    def unknown_entityref(self, ref): pass


# --------------------------------------------------------------------
# accelerated XML parser

class FastXMLParser:

    # Interface -- initialize and reset this instance
    def __init__(self, verbose=0):
	self.verbose = verbose
	self.reset()

    # Interface -- reset this instance.  Loses all unprocessed data
    def reset(self):
	self.rawdata = ''
	self.stack = []
	self.lasttag = '???'
	self.nomoretags = 0
	self.literal = 0
	self.lineno = 1
	self.parser = sgmlop.XMLParser()
	self.feed = self.parser.feed
	self.parser.register(self)

    # For derived classes only -- enter literal mode (CDATA) till EOF
    def setnomoretags(self):
	self.nomoretags = self.literal = 1

    # For derived classes only -- enter literal mode (CDATA)
    def setliteral(self, *args):
	self.literal = 1

    # Interface -- feed some data to the parser.  Call this as
    # often as you want, with as little or as much text as you
    # want (may include '\n').  (This just saves the text, all the
    # processing is done by goahead().)
    def feed(self, data): # overridden by reset
	self.parser.feed(data)

    # Interface -- handle the remaining data
    def close(self):
	try:
	    self.parser.close()
	finally:
	    self.parser = None

    # Interface -- translate references
    def translate_references(self, data):
	newdata = []
	i = 0
	while 1:
	    res = ref.search(data, i)
	    if res is None:
		newdata.append(data[i:])
		return string.join(newdata, '')
	    if data[res.end(0) - 1] != ';':
		self.syntax_error(self.lineno,
				  '; missing after entity/char reference')
	    newdata.append(data[i:res.start(0)])
	    str = res.group(1)
	    if str[0] == '#':
		if str[1] == 'x':
		    newdata.append(chr(string.atoi(str[2:], 16)))
		else:
		    newdata.append(chr(string.atoi(str[1:])))
	    else:
		try:
		    newdata.append(self.entitydefs[str])
		except KeyError:
		    # can't do it, so keep the entity ref in
		    newdata.append('&' + str + ';')
	    i = res.end(0)

    # Internal -- finish processing of start tag
    # Return -1 for unknown tag, 1 for balanced tag
    def finish_starttag(self, tag, data):
	# FIXME: should use the sgmlop attribute parser!!!
	attrs = {}
	k, j = 0, len(data)
	while k < j:
	    match = attrfind.match(data, k)
	    if not match: break
	    attrname, rest, attrvalue = match.group(1, 2, 3)
	    if not rest:
		attrvalue = attrname
	    elif attrvalue[:1] == '\'' == attrvalue[-1:] or \
		 attrvalue[:1] == '"' == attrvalue[-1:]:
		attrvalue = attrvalue[1:-1]
	    attrs[attrname] = attrvalue
	    k = match.end(0)
	# FIXME: should move this logic into sgmlop!
	self.stack.append(tag)
	try:
	    method = getattr(self, 'start_' + tag)
	except AttributeError:
	    self.unknown_starttag(tag, attrs)
	    return -1
	else:
	    self.handle_starttag(tag, method, attrs)
	    return 1

    # Internal -- finish processing of end tag
    def finish_endtag(self, tag):
	if not tag:
	    found = len(self.stack) - 1
	    if found < 0:
		self.unknown_endtag(tag)
		return
	else:
	    if tag not in self.stack:
		try:
		    method = getattr(self, 'end_' + tag)
		except AttributeError:
		    self.unknown_endtag(tag)
		return
	    found = len(self.stack)
	    for i in range(found):
		if self.stack[i] == tag: found = i
	while len(self.stack) > found:
	    tag = self.stack[-1]
	    try:
		method = getattr(self, 'end_' + tag)
	    except AttributeError:
		method = None
	    if method:
		self.handle_endtag(tag, method)
	    else:
		self.unknown_endtag(tag)
	    del self.stack[-1]

    # Overridable -- handle start tag
    def handle_starttag(self, tag, method, attrs):
	method(attrs)

    # Overridable -- handle end tag
    def handle_endtag(self, tag, method):
	method()

    # Example -- handle character reference, no need to override
    def handle_charref(self, name):
	try:
	    if name[0] == 'x':
		n = string.atoi(name[1:], 16)
	    else:
		n = string.atoi(name)
	except string.atoi_error:
	    self.unknown_charref(name)
	    return
	if not 0 <= n <= 255:
	    self.unknown_charref(name)
	    return
	self.handle_data(chr(n))

    # Definition of entities -- derived classes may override
    entitydefs = ENTITYDEFS

    # Example -- handle entity reference, no need to override
    def handle_entityref(self, name):
	table = self.entitydefs
	if table.has_key(name):
	    self.handle_data(table[name])
	else:
	    self.unknown_entityref(name)
	    return

    # Example -- handle data, should be overridden
    def handle_data(self, data):
	pass

    # Example -- handle cdata, could be overridden
    def handle_cdata(self, data):
	pass

    # Example -- handle comment, could be overridden
    def handle_comment(self, data):
	pass

    # Example -- handle processing instructions, could be overridden
    def handle_proc(self, name, data):
	pass

    # Example -- handle special instructions, could be overridden
    def handle_special(self, data):
	pass

    # Example -- handle relatively harmless syntax errors, could be overridden
    def syntax_error(self, lineno, message):
	raise RuntimeError, 'Syntax error at line %d: %s' % (lineno, message)

    # To be overridden -- handlers for unknown objects
    def unknown_starttag(self, tag, attrs): pass
    def unknown_endtag(self, tag): pass
    def unknown_charref(self, ref): pass
    def unknown_entityref(self, ref): pass


#sgmlop = None

# pick a suitable parser
if sgmlop:
    XMLParser = FastXMLParser
else:
    XMLParser = SlowXMLParser

# --------------------------------------------------------------------
# test stuff

class TestXMLParser(XMLParser):

    def __init__(self, verbose=0):
	self.testdata = ""
	XMLParser.__init__(self, verbose)

    def handle_data(self, data):
	self.testdata = self.testdata + data
	if len(`self.testdata`) >= 70:
	    self.flush()

    def flush(self):
	data = self.testdata
	if data:
	    self.testdata = ""
	    print 'data:', `data`

    def handle_cdata(self, data):
	self.flush()
	print 'cdata:', `data`

    def handle_proc(self, name, data):
	self.flush()
	print 'processing:',name,`data`

    def handle_special(self, data):
	self.flush()
	print 'special:',`data`

    def handle_comment(self, data):
	self.flush()
	r = `data`
	if len(r) > 68:
	    r = r[:32] + '...' + r[-32:]
	print 'comment:', r

    def syntax_error(self, lineno, message):
	print 'error at line %d:' % lineno, message

    def unknown_starttag(self, tag, attrs):
	self.flush()
	if not attrs:
	    print 'start tag: <' + tag + '>'
	else:
	    print 'start tag: <' + tag,
	    for name, value in attrs:
		print name + '=' + '"' + value + '"',
	    print '>'

    def unknown_endtag(self, tag):
	self.flush()
	print 'end tag: </' + tag + '>'

    def unknown_entityref(self, ref):
	self.flush()
	print '*** unknown entity ref: &' + ref + ';'

    def unknown_charref(self, ref):
	self.flush()
	print '*** unknown char ref: &#' + ref + ';'

    def close(self):
	XMLParser.close(self)
	self.flush()

def test(args = None):
    import sys

    if not args:
	args = sys.argv[1:]

    if args and args[0] == '-s':
	args = args[1:]
	klass = XMLParser
    else:
	klass = TestXMLParser

    if args:
	file = args[0]
    else:
	file = 'test.xml'

    if file == '-':
	f = sys.stdin
    else:
	try:
	    f = open(file, 'r')
	except IOError, msg:
	    print file, ":", msg
	    sys.exit(1)

    data = f.read()
    if f is not sys.stdin:
	f.close()

    x = klass()
    for c in data:
	x.feed(c)
    x.close()


if __name__ == '__main__':
    test()




"""These are the DTD-aware classes of xmlproc. They provide both the
DTD event consumers for the DTD parser as well as the objects that
store DTD information for retrieval by clients (including the
validating parser)."""

import types

from xmlutils import *
from xmlapp import *

# ==============================
# WFC-DTD
# ==============================

class WFCDTD(DTDConsumer):
    "DTD-representing class for the WFC parser."

    def __init__(self,err):
	DTDConsumer.__init__(self,err)
	self.gen_ents={}
	self.param_ents={}
	self.elems={}

        self.dtd_listener=DTDConsumer(err)
        
	# Adding predefined entities
	for name in predef_ents.keys():
	    self.new_general_entity(name,predef_ents[name])

    def set_dtd_listener(self,listener):
        "Registers an object that listens for DTD parse events."
        self.dtd_listener=listener
            
    def resolve_pe(self,name):
	"""Returns the entitiy object associated with this parameter entity
        name. Throws KeyError if the entity is not declared."""
        return self.param_ents[name]

    def resolve_ge(self,name):
	"""Returns the entitiy object associated with this general entity
        name. Throws KeyError if the entity is not declared."""
        return self.gen_ents[name]

    def get_elem(self,name):
	"""Returns the declaration of this element. Throws KeyError if the
	element does not exist."""
	return self.elems[name]

    def new_attribute(self,elem,attr,a_type,a_decl,a_def):
	"Receives the declaration of a new attribute."
        self.dtd_listener.new_attribute(elem,attr,a_type,a_decl,a_def)
        
        if not self.elems.has_key(elem):
	    self.elems[elem]=ElementTypeAny(elem) # Adding dummy

        self.elems[elem].add_attr(attr,a_type,a_decl,a_def,self.err)
                
    # --- Echoing DTD parse events
    
    def new_general_entity(self,name,val):
        ent=InternalEntity(name,val)
	self.gen_ents[ent.name]=ent
        self.dtd_listener.new_general_entity(name,val)

    def new_parameter_entity(self,name,val):
        ent=InternalEntity(name,val)
	self.param_ents[ent.name]=ent
        self.dtd_listener.new_parameter_entity(name,val)

    def new_external_entity(self,ent_name,sys_id,pub_id,ndata):
        ent=ExternalEntity(ent_name,sys_id,pub_id,ndata)
	self.gen_ents[ent.name]=ent
        self.dtd_listener.new_external_entity(ent_name,sys_id,pub_id,ndata)

    def new_external_pe(self,name,sysid,pubid):
        ent=ExternalEntity(name,sysid,pubid,"")
	self.param_ents[ent.name]=ent
        self.dtd_listener.new_external_pe(name,sysid,pubid)
	
    def dtd_end(self):
        self.dtd_listener.dtd_end()
    
    def new_comment(self,contents):
        self.dtd_listener.new_comment(contents)

    def new_pi(self,target,rem):
        self.dtd_listener.new_pi(target,rem)
    
    def new_notation(self,name,sysid,pubid):
        self.dtd_listener.new_notation(name,sysid,pubid)

    def new_element_type(self,elem_name,elem_cont):
        self.dtd_listener.new_element_type(elem_name,elem_cont)
    
# ==============================
# DTD consumer for the validating parser
# ==============================
    
class CompleteDTD(WFCDTD):
    "Complete DTD handler for the validating parser."

    def __init__(self,err):
	WFCDTD.__init__(self,err)
	self.notations={}
	self.attlists={}  # Attribute lists of elements not yet declared
	self.root_elem=None
	
    def get_root_elem(self):
	"Returns the name of the declared root element."
	return self.root_elem

    def get_notation(self,name):
	"""Returns the declaration of the notation. Throws KeyError if the
        notation does not exist."""
        return self.notations[name]

    # --- DTD parse events

    def dtd_end(self):
	for elem in self.attlists.keys():
	    self.err.warning("Element '%s' has attribute list, but no element "
			     "declaration" % elem)

	del self.attlists  # Not needed any more, can free this memory
        self.dtd_listener.dtd_end()
	
    def new_notation(self,name,sysid,pubid):
	self.notations[name]=(sysid,pubid)
        self.dtd_listener.new_notation(name,sysid,pubid)

    def new_element_type(self,elem_name,elem_cont):
	if self.elems.has_key(elem_name):
	    self.err.fatal("Element '%s' declared twice" % elem_name)

	if elem_cont==None:
	    model=make_empty_model()
	elif elem_cont!=1: 
	    model=fnda2fda(elem_cont.get_states(),self.err)

	if elem_cont==1:
	    self.elems[elem_name]=ElementTypeAny(elem_name)
	else:
	    self.elems[elem_name]=ElementType(elem_name,model)

	if self.attlists.has_key(elem_name):
	    for (attr,a_type,a_decl,a_def) in self.attlists[elem_name]:
		self.elems[elem_name].add_attr(attr,a_type,a_decl,a_def,\
					       self.err)
	    del self.attlists[elem_name]
            
        self.dtd_listener.new_element_type(elem_name,elem_cont)
	        
    def new_attribute(self,elem,attr,a_type,a_decl,a_def):
	"Receives the declaration of a new attribute."
        self.dtd_listener.new_attribute(elem,attr,a_type,a_decl,a_def)
	try:
	    self.elems[elem].add_attr(attr,a_type,a_decl,a_def,self.err)
	except KeyError,e:
	    try:
		self.attlists[elem].append((attr,a_type,a_decl,a_def))
	    except KeyError,e:
		self.attlists[elem]=[(attr,a_type,a_decl,a_def)]
                
# ==============================
# Represents an XML element type
# ==============================
    
class ElementType:
    "Represents an element type."

    def __init__(self,name,cont):
	self.name=name
	self.attrhash={}
	self.content_model=cont

    def get_name(self):
	"Returns the name of the element type."
	return self.name
	
    def get_attr_list(self):
	"Returns a list of the declared attribute names."
	return self.attrhash.keys()
	
    def get_attr(self,name):
	"Returns the attribute or throws a KeyError if it's not declared."
	return self.attrhash[name]
	
    def add_attr(self,attr,a_type,a_decl,a_def,err):
	"Adds a new attribute to the element."
	if self.attrhash.has_key(attr):
	    err.warning("Attribute '%s' defined more than once" % attr)
	
	if a_type=="ID":
	    for attr_name in self.attrhash.keys():
		if self.attrhash[attr_name].type=="ID":
		    err.error("Only one ID attribute allowed on each element "
			      "type")

	    if a_decl!="#REQUIRED" and a_decl!="#IMPLIED":
		err.error("ID attributes cannot be #FIXED")
	
	self.attrhash[attr]=Attribute(attr,a_type,a_decl,a_def)
	
    def get_start_state(self):
	"Return the start state of this content model."
	return self.content_model["start"]
	
    def final_state(self,state):
	"True if 'state' is a final state."
	return self.content_model["final"]<=state
	
    def next_state(self,state,elem_name):
	"""Returns the next state of the content model from the given one
        when elem_name is encountered. Character data is represented as
        '#PCDATA'. If 0 is returned the element is not allowed here."""
	
	for (to,trans_name) in self.content_model[state]:
	    if trans_name==elem_name:
		return to

	return 0

# --- Element types with ANY content

class ElementTypeAny(ElementType):

    def __init__(self,name):
	ElementType.__init__(self,name,None)

    def get_start_state(self):
	return 1

    def final_state(self,state):
	return 1

    def next_state(self,state,elem_name):
	return 1
    
# ==============================
# Attribute
# ==============================

class Attribute:
    "Represents a declared attribute."

    def __init__(self,name,type,decl,default):
	self.name=name
	self.type=type
	self.decl=decl
	self.default=default

        # Handling code for special attribute xml:space
        
        if name=="xml:space":
            if type(self.type)==types.StringType:
                err.error("xml:space must be declared an enumeration type")
                return

            if len(self.type)!=2:
                error=1
            else:
                if (self.type[0]=="default" and self.type[1]=="preserve") or \
                   (self.type[1]=="default" and self.type[0]=="preserve"):
                    error=0
                else:
                    error=1

            if error:
                err.error("xml:space must have exactly the values 'default'"
                          " and 'preserve'")                            
                
    def validate(self,value,err):
	"Validates given value for correctness."

	if type(self.type)!=types.StringType:
	    for val in self.type:
		if val==value: return
	    err.error("'%s' is not an allowed value for the '%s' attribute"\
		      % (value,self.name))
	elif self.type=="CDATA":
	    return
	elif self.type=="ID" or self.type=="IDREF":
	    if not matches(reg_name,value):
		err.error("Value of '%s' attribute must be a valid name" %\
			  self.name)
	elif self.type=="NMTOKEN":
	    if not matches(reg_nmtoken,value):
		err.error("Value of '%s' attribute not a valid name token" %\
			  self.name)
	elif self.type=="NMTOKENS":
	    if not matches(reg_nmtokens,value):
		err.error("Value of '%s' attribute not a valid name token" \
			  " sequence" % self.name)

    def get_name(self):
        "Returns the attribute name."
        return self.name
        
    def get_type(self):
        "Returns the type of the attribute. (ID, CDATA etc)"
        return self.type

    def get_decl(self):
        "Returns the declaration (#IMPLIED, #REQUIRED, #FIXED or #DEFAULT)."
        return self.decl

    def get_default(self):
        """Returns the default value of the attribute, or None if none has
        been declared."""
        return self.default    
                
# ==============================
# Entities
# ==============================

class InternalEntity:

    def __init__(self,name,value):
	self.name=name
	self.value=value

    def is_internal(self):
	return 1

class ExternalEntity:

    def __init__(self,name,pubid,sysid,notation):
	self.name=name
	self.pubid=pubid
	self.sysid=sysid
	self.notation=notation

    def is_parsed(self):
        "True if this is a parsed entity."
	return self.notation==""
	
    def is_internal(self):
	return 0

    def get_pubid(self):
        "Returns the public identifier of the entity."
        return self.pubid

    def get_sysid(self):
        "Returns the system identifier of the entity."
        return self.sysid
    

# ==============================
# Internal classes
# ==============================
    
# Content model class

class ContentModel:
    "Represents a content model. (Internal.)"

    def __init__(self,contents,modifier):
	self.contents=contents
	self.modifier=modifier

    def __getitem__(self,itemno):
	"Returns the requested content particle from the content model."
	return self.contents[itemno]

    def get_states(self):
	if type(self[0][0])==types.InstanceType:
	    transitions=self[0][0].get_states()
	else:
	    transitions=[[(1,self[0][0])],[]]

	first=0
	last=len(transitions)-1
	if self.modifier=="?":
	    transitions[first].append((last,""))
	elif self.modifier=="+":
	    transitions[last].append((first,""))	    
	elif self.modifier=="*":
	    transitions[first].append((last,""))
	    transitions[last].append((first,""))
	
	return transitions

# Sequential content model
    
class SeqContentModel(ContentModel):
    "Represents a sequential content model. (Internal.)"
    
    def get_states(cont_list):
	"Sketch of a function that creates a state model from the content\
	model."

	transitions=[[]]
	cur_state=0

	for (cp,mod) in cont_list:
	    orig_state=cur_state
	    if types.InstanceType==type(cp):
		for newstate in cp.get_states():
		    for (to,what) in newstate:
			transitions[cur_state].append((to+orig_state,what))
		    cur_state=cur_state+1
		    transitions.append([])
		transitions=transitions[:-1]
		cur_state=cur_state-1  # Siste må smeltes inn med gammel
	    else:
		cur_state=cur_state+1
		transitions.append([])
		transitions[orig_state].append((cur_state,cp))

	    if mod=="?":
		transitions[orig_state].append((cur_state,""))
	    elif mod=="+":
		transitions[cur_state].append((orig_state,""))
	    elif mod=="*":
		transitions[orig_state].append((cur_state,""))
		transitions[cur_state].append((orig_state,""))

	first=0
	last=len(transitions)-1
	if cont_list.modifier=="?":
	    transitions[first].append((last,""))
	elif cont_list.modifier=="+":
	    transitions[last].append((first,""))	    
	elif cont_list.modifier=="*":
	    transitions[first].append((last,""))
	    transitions[last].append((first,""))
	
	return transitions

# Choice content model

class ChoiceContentModel(ContentModel):
    "Represents a choice content model. (Internal.)"

    def get_states(cont_list):
	"Sketch of a function that creates a state model from the content\
	model."

	transitions=[[],[]]
	for (cp,mod) in cont_list:
	    transitions[0].append((1,cp))

	first=0
	last=len(transitions)-1
	if cont_list.modifier=="?":
	    transitions[first].append((last,""))
	elif cont_list.modifier=="+":
	    transitions[last].append((first,""))	    
	elif cont_list.modifier=="*":
	    transitions[first].append((last,""))
	    transitions[last].append((first,""))

	return transitions
    
# ==============================
# Conversion of FDAs
# ==============================

def hash(included):
    "Creates a hash number from the included array."
    no=0
    exp=1
    for state in included:
	if state:
	    no=no+exp
	exp=exp*2

    return no

def fnda2fda(transitions,err):
    """Converts a finite-state non-deterministic automaton into a deterministic
    one."""

    transitions.append([])
    new_states={}

    # Compute the e-closure of the start state
    included=[0]*len(transitions)
    compute_closure(0,included,transitions)
    state_key=hash(included)

    # Add transitions and the other states
    add_transitions(0,transitions,new_states,included,state_key,err)

    for state in new_states.keys():
	if state % 2==1:
	    new_states["start"]=state
	    break

    new_states["final"]=pow(2,len(transitions)-2) # 2 because we add []

    return new_states
    
def add_transitions(ix,transitions,new_states,included,state_key,err):
    "Set up transitions and create new states."

    if new_states.has_key(state_key):
	return                   # We've been here before
    else:
	new_states[state_key]=[] # OK, a new one, create it
    
    no=0
    for old_state in included:
	if old_state:
	    for (to,what) in transitions[no]:
		if what!="":
		    new_inc=[0]*len(transitions)
		    compute_closure(to,new_inc,transitions)
		    new_state=hash(new_inc)

		    for (to,over) in new_states[state_key]:
			if over==what:
			    err.warning("Ambiguous content model")
		    
		    new_states[state_key].append(new_state,what)
		    add_transitions(to,transitions,new_states,new_inc,\
				    new_state,err)

	no=no+1

    
def compute_closure(ix,included,transitions):
    "Computes the e-closure of this state."
    if included[ix]: return
    
    included[ix]=1
    for (to,what) in transitions[ix]:
	if what=="": compute_closure(to,included,transitions)

def print_states(states):
    assert not (states.has_key("start") or states.has_key("final"))
    
    for trans_key in states.keys():
	trans=states[trans_key]
	print "State: "+`trans_key`
	for (to,what) in trans:
	    try:
		print "  To: "+`to`+" over: "+what
	    except TypeError,e:
		print "ERROR: "+`what`

    raw_input()

def make_empty_model():
    "Constructs a state model for empty content models."
    return { 1:[], "final":1, "start":1 }



'''DOM builder from ESIS output.'''

# Hacked from ``pygrove.py'' from Paul Prescod.

import string, re, regsub, sys
from xml.dom.core import *
from xml.dom.builder import Builder


_sdata_dict = {
	'acirc': 'â',
	'agrave': 'à',
	'ccedil': 'ç',
	'dollar': '$',
	'eacute': 'é',
	'ecirc': 'ê',
	'egrave': 'è',
	'icirc': 'î',
	'ocirc': 'ô',
	'percnt': '%',
	'ugrave': 'ù',
}

def handle_sdata(sdata):
	return _sdata_dict.get(sdata, 'unknown')


class EsisBuilder(Builder):
	
	def __init__(self):
		Builder.__init__(self)
		self.attr_store = {}
		#self.sdata_handler = handle_sdata

	def feed(self, data):
		for line in string.split(data, '\n'):
			if not line: 
				break
			event = line[0]
			text = line[1:]

			if event == '(':
				element = self.dom_factory.createElement(text, self.attr_store)
				self.attr_store = {}
				self.push(element)

			elif event == ')':
				self.pop()

			elif event == 'A':
				l = re.split(' ', text, 2)
				name = l[0]
				value = ESISDecode(l[2])
				self.attr_store[name] = value

			elif event == '-':
				text = self.dom_factory.createTextNode(ESISDecode(text))
				self.push(text)

			elif event == 'C':
				return

			else:
				sys.stderr.write('Unknow event: ' + `line` + '\n')


backslash = r"\\"
regor = "|"

find = "(" + r"\\\\" + regor \
		+ r"\\n" + regor \
		+ r"\\\|\[[^" +backslash+ "]*" + r"\]\\\|" + regor \
		+ r"\\[0-9]+" + regor \
		+ r"\\#[0-9]+" + regor \
		+ r"\\%[0-9]+" + regor \
		+ ")"

def fix(text):
	if (not text) or (text[0] != "\\"):
		return text
	text = text[1:]
	if (text == "\\"):
		return "\\"
	elif (text == "n"):
		return "\n"
	elif(text[0] == "|"):
		return handle_sdata(string.strip(text[2:-3]))
		#return '&' + string.strip(text[2:-3]) + ';'
	elif(text[0] == "#"):
		return chr(string.atoi(text[1:],10))
	elif(text[0] == "%"):
		return chr(string.atoi(text[1:],10))
	else:
		return chr(string.atoi(text[0:3], 8)) + text[3:]

def ESISEncode(text):
	return re.sub("\n", "\\n", text);

def ESISDecode(text):
	#return regsub.gsub("\\\\n", "\n", text)

	#print `text`
	prog = re.compile(find)
	parts = prog.split(text, find)
	#print parts
	parts = map(fix, parts)
	res = string.join(parts, "");
	#print res
	return res


if __name__ == '__main__':
	import sys
	from xml.dom.writer import XmlLineariser

	p = EsisBuilder()
	p.feed(open(sys.argv[1]).read())

	w = XmlLineariser()
	w.add_newline_after = [ 'p', 'title', 'abstract' ]
	print w.linearise(p.document.documentElement)

# vim:ts=2:ai



'''
sax_builder.py: SaxBuilder class implementation.

SaxBuilder builds a DOM-conformant grove from evenements generated by a
SAX-conformant driver.
'''

from xml.sax.saxlib import HandlerBase
from builder import Builder


class SaxBuilder(Builder, HandlerBase):
	def __init__(self):
		HandlerBase.__init__(self)
		Builder.__init__(self)

	def characters(self, char, start, length):
		self.text(char[start:start+length])


# Test.
if __name__ == '__main__':
	print 'Testing dom.sax_builder. Usage: python dom/sax_builder <xml_file>.'

	import saxlib.xlsax, sys, writer

	p = saxlib.xlsax.SAX_XMLParser()
	dh = SaxBuilder()
	p.setDocumentHandler(dh)
	p.parse('', sys.argv[1])

	w = writer.XmlLineariser()
	print w.linearise(dh.document)
	



'''HTML parser, built from standard lib's sgmllib.

Tag names are normalised to upper case, the usual HTML fashion.
'''

from sgmllib import SGMLParser
from xml.dom.core import DOMFactory
from xml.dom.builder import Builder
import string


class HtmlBuilder(SGMLParser, Builder):
	from htmlentitydefs import entitydefs
	
	def __init__(self):
		SGMLParser.__init__(self)
		Builder.__init__(self)

		self.empties = [
			'META', 'BASE', 'LINK', 
			'HR', 'BR',
			'IMG', 'PARAM',
			'INPUT', 'OPTION', 'ISINDEX'
		]
		list = ('OL', 'UL', 'DL')
		heading = ('H1', 'H2', 'H3', 'H4', 'H5', 'H6')
		blocks = ('P', 'ADDRESS', 'BLOCKQUOTE', 'FORM', 'TABLE', 'PRE') + \
			heading # + list
		self.infer_ends = {
			'P': blocks,

			'LI': ('LI',),
			'DT': ('DT',),
			'DD': ('DT', 'DD'),

			'TR': ('TR',), 
			'TH': ('TH', 'TD', 'TR'),
			'TD': ('TH', 'TD', 'TR'),
		}

	
	def unknown_starttag(self, tag, attrs):
		tag = string.upper(tag)
		#print 'starting', tag
		attributes = {}
		for k, v in attrs:
			attributes[string.upper(k)] = v

		#print self.stack
		while self.stack:
			if self.infer_ends.has_key(self.stack[-1]): 
				if tag in self.infer_ends[self.stack[-1]]:
					#print tag, 'ending', self.stack[-1]
					Builder.endElement(self, self.stack[-1])
					del self.stack[-1]
					#print self.stack
				else:
					break
			else:
				break
		#print self.stack

		Builder.startElement(self, tag, attributes)
		if not tag in self.empties:
			self.stack.append(tag)
		else:
			Builder.endElement(self, tag)


	def unknown_endtag(self, tag):
		tag = string.upper(tag)
		#print 'ending', tag

		while self.stack:
			if tag in self.empties:
				continue
			start_tag = self.stack[-1]
			del self.stack[-1]
			Builder.endElement(self, start_tag)
			if start_tag == tag:
				break

	def handle_data(self, s):
		#print `s`
		Builder.text(self, s)


# Test.
if __name__ == '__main__':
	import sys
	b = HtmlBuilder()
	b.feed(open(sys.argv[1]).read())
	b.close()
	#print b.document
	#print b.document.documentElement

	from writer import HtmlLineariser
	w = HtmlLineariser()
	print w.linearise(b.document)




"""A Python translation of the SAX parser API. This file provides only
default classes with absolutely minimum functionality, from which
drivers and applications can be subclassed.

Many of these classes are empty and are included only as documentation
of the interfaces."""

# --- Locator

class Locator:
    """Interface for associating a SAX event with a document
    location. A locator object will return valid results only during
    calls to SAXDocumentHandler methods; at any other time, the
    results are unpredictable."""
    
    def getColumnNumber(self):
	"Return the column number where the current event ends."
	return -1

    def getLineNumber(self):
	"Return the line number where the current event ends."
	return -1

    def getPublicId(self):
	"Return the public identifier for the current event."
	return ""

    def getSystemId(self):
	"Return the system identifier for the current event."
	return ""

# --- SAXException

class SAXException(Exception):
    """Encapsulate an XML error or warning. This class can contain
    basic error or warning information from either the XML parser or
    the application: you can subclass it to provide additional
    functionality, or to add localization. Note that although you will
    receive a SAXException as the argument to the handlers in the
    ErrorHandler interface, you are not actually required to throw
    the exception; instead, you can simply read the information in
    it."""
    
    def __init__(self, msg, exception):
        """Creates an exception. The message is required, but the exception
        can be None."""
        self.msg=msg
        self.exception=exception

    def getMessage(self):
	"Return a message for this exception."
	return self.msg

    def getException(self):
        "Return the embedded exception, if any."
        return self.exception

    def __str__(self):
        "Create a string representation of the exception."
        return self.msg

# --- SAXParseException

class SAXParseException(SAXException):    
    """Encapsulate an XML parse error or warning.
    
    This exception will include information for locating the error in
    the original XML document. Note that although the application will
    receive a SAXParseException as the argument to the handlers in the
    ErrorHandler interface, the application is not actually required
    to throw the exception; instead, it can simply read the
    information in it and take a different action.

    Since this exception is a subclass of SAXException, it inherits
    the ability to wrap another exception."""
    
    def __init__(self, msg, exception, locator):
        "Creates the exception. The exception parameter is allowed to be None."
        SAXException.__init__(self,msg,exception)
        self.locator=locator
        
    def getColumnNumber(self):
	"""The column number of the end of the text where the exception
	occurred."""
	return self.locator.getColumnNumber()

    def getLineNumber(self):
	"The line number of the end of the text where the exception occurred."
	return self.locator.getLineNumber()

    def getPublicId(self):
	"Get the public identifier of the entity where the exception occurred."
	return self.locator.getPublicId()

    def getSystemId(self):
	"Get the system identifier of the entity where the exception occurred."
	return self.locator.getSystemId()

    def __str__(self):
        "Create a string representation of the exception."
        return "%s at %s:%d:%d" % (self.msg,self.getSystemId(),
                                   self.getColumnNumber(),self.getLineNumber())
    
# --- EntityResolver

class EntityResolver:
    """Basic interface for resolving entities. If you create an object
    implementing this interface, then register the object with your
    Parser, the parser will call the method in your object to
    resolve all external entities. Note that HandlerBase implements
    this interface with the default behaviour."""
    
    def resolveEntity(self, publicId, systemId):
	"Resolve the system identifier of an entity."
	return systemId

# --- ErrorHandler

class ErrorHandler:
    """Basic interface for SAX error handlers. If you create an object
    that implements this interface, then register the object with your
    Parser, the parser will call the methods in your object to report
    all warnings and errors. There are three levels of errors
    available: warnings, (possibly) recoverable errors, and
    unrecoverable errors. All methods take a SAXParseException as the
    only parameter."""

    def error(self, exception):
	"Handle a recoverable error."
	pass

    def fatalError(self, exception):
	"Handle a non-recoverable error."
	pass

    def warning(self, exception):
	"Handle a warning."
	pass

# --- AttributeList

class AttributeList:
    """Interface for an attribute list. This interface provides
    information about a list of attributes for an element (only
    specified or defaulted attributes will be reported). Note that the
    information returned by this object will be valid only during the
    scope of the DocumentHandler.startElement callback, and the
    attributes will not necessarily be provided in the order declared
    or specified."""

    def getLength(self):
	"Return the number of attributes in list."
        pass

    def getName(self, i):
	"Return the name of an attribute in the list."
        pass

    def getType(self, i):
	"""Return the type of an attribute in the list. (Parameter can be
	either integer index or attribute name.)"""
        pass

    def getValue(self, i):
	"""Return the value of an attribute in the list. (Parameter can be
	either integer index or attribute name.)"""
        pass

    def __len__(self):
	"Alias for getLength."
        pass

    def __getitem__(self, key):
	"Alias for getName (if key is an integer) and getValue (if string)."
        pass

    def keys(self):
	"Returns a list of the attribute names."
        pass

    def has_key(self, key):
	"True if the attribute is in the list, false otherwise."
        pass

# --- DTDHandler

class DTDHandler:
    """Handle DTD events. This interface specifies only those DTD
    events required for basic parsing (unparsed entities and
    attributes). If you do not want to implement the entire interface,
    you can extend HandlerBase, which implements the default
    behaviour."""

    def notationDecl(self, name, publicId, systemId):
	"Handle a notation declaration event."
	pass

    def unparsedEntityDecl(self, name, publicId, systemId, ndata):
	"Handle an unparsed entity declaration event."
	pass

# --- DocumentHandler

class DocumentHandler:
    """Handle general document events. This is the main client
    interface for SAX: it contains callbacks for the most important
    document events, such as the start and end of elements. You need
    to create an object that implements this interface, and then
    register it with the Parser. If you do not want to implement
    the entire interface, you can derive a class from HandlerBase,
    which implements the default functionality. You can find the
    location of any document event using the Locator interface
    supplied by setDocumentLocator()."""

    def characters(self, ch, start, length):
	"Handle a character data event."
	pass

    def endDocument(self):
	"Handle an event for the end of a document."
	pass

    def endElement(self, name):
	"Handle an event for the end of an element."
	pass

    def ignorableWhitespace(self, ch, start, length):
	"Handle an event for ignorable whitespace in element content."
	pass

    def processingInstruction(self, target, data):
	"Handle a processing instruction event."
	pass

    def setDocumentLocator(self, locator):
	"Receive an object for locating the origin of SAX document events."
	pass

    def startDocument(self):
	"Handle an event for the beginning of a document."
	pass

    def startElement(self, name, atts):
	"Handle an event for the beginning of an element."
	pass

# --- Parser

class Parser:
    """Basic interface for SAX (Simple API for XML) parsers. All SAX
    parsers must implement this basic interface: it allows users to
    register handlers for different types of events and to initiate a
    parse from a URI, a character stream, or a byte stream. SAX
    parsers should also implement a zero-argument constructor."""

    def __init__(self):
	self.doc_handler=DocumentHandler()
	self.dtd_handler=DTDHandler()
	self.ent_handler=EntityResolver()
	self.err_handler=ErrorHandler()

    def parse(self, systemId):
	"Parse an XML document from a system identifier."
	pass

    def parseFile(self, fileobj):
        "Parse an XML document from a file-like object."
        pass

    def setDocumentHandler(self, handler):
	"Register an object to receive basic document-related events."
	self.doc_handler=handler

    def setDTDHandler(self, handler):
	"Register an object to receive basic DTD-related events."
	self.dtd_handler=handler

    def setEntityResolver(self, resolver):
	"Register an object to resolve external entities."
	self.ent_handler=resolver

    def setErrorHandler(self, handler):
	"Register an object to receive error-message events."
	self.err_handler=handler

    def setLocale(self, locale):
        """Allow an application to set the locale for errors and warnings. 
   
        SAX parsers are not required to provide localisation for errors
        and warnings; if they cannot support the requested locale,
        however, they must throw a SAX exception. Applications may
        request a locale change in the middle of a parse."""
        raise SAXException("Locale support not implemented",None)

# --- HandlerBase

class HandlerBase(EntityResolver, DTDHandler, DocumentHandler,\
		     ErrorHandler):
    """Default base class for handlers. This class implements the
    default behaviour for four SAX interfaces: EntityResolver,
    DTDHandler, DocumentHandler, and ErrorHandler: rather
    than implementing those full interfaces, you may simply extend
    this class and override the methods that you need. Note that the
    use of this class is optional (you are free to implement the
    interfaces directly if you wish)."""

    def __init__(self):
        pass



# A parser for SGML, using the derived class as static DTD.

# XXX This only supports those SGML features used by HTML.

# XXX There should be a way to distinguish between PCDATA (parsed
# character data -- the normal case), RCDATA (replaceable character
# data -- only char and entity references and end tags are special)
# and CDATA (character data -- only end tags are special).

# sgmlop support added by fredrik@pythonware.com (April 6, 1998)

import re
import string

try:
    import sgmlop
except ImportError:
    sgmlop = None

# standard entity defs

ENTITYDEFS = {
    'lt': '<',
    'gt': '>',
    'amp': '&',
    'quot': '"',
    'apos': '\''
    }

# SGML parser base class -- find tags and call handler functions.
# Usage: p = SGMLParser(); p.feed(data); ...; p.close().
# The dtd is defined by deriving a class which defines methods
# with special names to handle tags: start_foo and end_foo to handle
# <foo> and </foo>, respectively, or do_foo to handle <foo> by itself.
# (Tags are converted to lower case for this purpose.)  The data
# between tags is passed to the parser by calling self.handle_data()
# with some data as argument (the data may be split up in arbutrary
# chunks).  Entity references are passed by calling
# self.handle_entityref() with the entity reference as argument.

# --------------------------------------------------------------------
# original re-based SGML parser

interesting = re.compile('[&<]')
incomplete = re.compile('&([a-zA-Z][a-zA-Z0-9]*|#[0-9]*)?|'
			   '<([a-zA-Z][^<>]*|'
			      '/([a-zA-Z][^<>]*)?|'
			      '![^<>]*)?')

entityref = re.compile('&([a-zA-Z][a-zA-Z0-9]*)[^a-zA-Z0-9]')
charref = re.compile('&#([0-9]+)[^0-9]')

starttagopen = re.compile('<[>a-zA-Z]')
shorttagopen = re.compile('<[a-zA-Z][a-zA-Z0-9]*/')
shorttag = re.compile('<([a-zA-Z][a-zA-Z0-9]*)/([^/]*)/')
endtagopen = re.compile('</[<>a-zA-Z]')
endbracket = re.compile('[<>]')
special = re.compile('<![^<>]*>')
commentopen = re.compile('<!--')
commentclose = re.compile('--[ \t\n]*>')
tagfind = re.compile('[a-zA-Z][a-zA-Z0-9]*')
attrfind = re.compile(
    '[ \t\n]*([a-zA-Z_][-.a-zA-Z_0-9]*)'
    '([ \t\n]*=[ \t\n]*'
    r'(\'[^\']*\'|"[^"]*"|[-a-zA-Z0-9./:+*%?!\(\)_#=~]*))?')


class SlowSGMLParser:

    # Interface -- initialize and reset this instance
    def __init__(self, verbose=0):
	self.verbose = verbose
	self.reset()

    # Interface -- reset this instance.  Loses all unprocessed data
    def reset(self):
	self.rawdata = ''
	self.stack = []
	self.lasttag = '???'
	self.nomoretags = 0
	self.literal = 0

    # For derived classes only -- enter literal mode (CDATA) till EOF
    def setnomoretags(self):
	self.nomoretags = self.literal = 1

    # For derived classes only -- enter literal mode (CDATA)
    def setliteral(self, *args):
	self.literal = 1

    # Interface -- feed some data to the parser.  Call this as
    # often as you want, with as little or as much text as you
    # want (may include '\n').  (This just saves the text, all the
    # processing is done by goahead().)
    def feed(self, data):
	self.rawdata = self.rawdata + data
	self.goahead(0)

    # Interface -- handle the remaining data
    def close(self):
	self.goahead(1)

    # Internal -- handle data as far as reasonable.  May leave state
    # and data to be processed by a subsequent call.  If 'end' is
    # true, force handling all data as if followed by EOF marker.
    def goahead(self, end):
	rawdata = self.rawdata
	i = 0
	n = len(rawdata)
	while i < n:
	    if self.nomoretags:
		self.handle_data(rawdata[i:n])
		i = n
		break
	    match = interesting.search(rawdata, i)
	    if match: j = match.start(0)
	    else: j = n
	    if i < j: self.handle_data(rawdata[i:j])
	    i = j
	    if i == n: break
	    if rawdata[i] == '<':
		if starttagopen.match(rawdata, i):
		    if self.literal:
			self.handle_data(rawdata[i])
			i = i+1
			continue
		    k = self.parse_starttag(i)
		    if k < 0: break
		    i = k
		    continue
		if endtagopen.match(rawdata, i):
		    k = self.parse_endtag(i)
		    if k < 0: break
		    i =  k
		    self.literal = 0
		    continue
		if commentopen.match(rawdata, i):
		    if self.literal:
			self.handle_data(rawdata[i])
			i = i+1
			continue
		    k = self.parse_comment(i)
		    if k < 0: break
		    i = i+k
		    continue
		match = special.match(rawdata, i)
		if match:
		    if self.literal:
			self.handle_data(rawdata[i])
			i = i+1
			continue
		    i = match.end(0)
		    continue
	    elif rawdata[i] == '&':
		match = charref.match(rawdata, i)
		if match:
		    name = match.group(1)
		    self.handle_charref(name)
		    i = match.end(0)
		    if rawdata[i-1] != ';': i = i-1
		    continue
		match = entityref.match(rawdata, i)
		if match:
		    name = match.group(1)
		    self.handle_entityref(name)
		    i = match.end(0)
		    if rawdata[i-1] != ';': i = i-1
		    continue
	    else:
		raise RuntimeError, 'neither < nor & ??'
	    # We get here only if incomplete matches but
	    # nothing else
	    match = incomplete.match(rawdata, i)
	    if not match:
		self.handle_data(rawdata[i])
		i = i+1
		continue
	    j = match.end(0)
	    if j == n:
		break # Really incomplete
	    self.handle_data(rawdata[i:j])
	    i = j
	# end while
	if end and i < n:
	    self.handle_data(rawdata[i:n])
	    i = n
	self.rawdata = rawdata[i:]
	# XXX if end: check for empty stack

    # Internal -- parse comment, return length or -1 if not terminated
    def parse_comment(self, i):
	rawdata = self.rawdata
	if rawdata[i:i+4] <> '<!--':
	    raise RuntimeError, 'unexpected call to handle_comment'
	match = commentclose.search(rawdata, i+4)
	if not match:
	    return -1
	j = match.start(0)
	self.handle_comment(rawdata[i+4: j])
	j = match.end(0)
	return j-i

    # Internal -- handle starttag, return length or -1 if not terminated
    def parse_starttag(self, i):
	rawdata = self.rawdata
	if shorttagopen.match(rawdata, i):
	    # SGML shorthand: <tag/data/ == <tag>data</tag>
	    # XXX Can data contain &... (entity or char refs)?
	    # XXX Can data contain < or > (tag characters)?
	    # XXX Can there be whitespace before the first /?
	    match = shorttag.match(rawdata, i)
	    if not match:
		return -1
	    tag, data = match.group(1, 2)
	    tag = string.lower(tag)
	    self.finish_shorttag(tag, data)
	    k = match.end(0)
	    return k
	# XXX The following should skip matching quotes (' or ")
	match = endbracket.search(rawdata, i+1)
	if not match:
	    return -1
	j = match.start(0)
	# Now parse the data between i+1 and j into a tag and attrs
	attrs = []
	if rawdata[i:i+2] == '<>':
	    # SGML shorthand: <> == <last open tag seen>
	    k = j
	    tag = self.lasttag
	else:
	    match = tagfind.match(rawdata, i+1)
	    if not match:
		raise RuntimeError, 'unexpected call to parse_starttag'
	    k = match.end(0)
	    tag = string.lower(rawdata[i+1:k])
	    self.lasttag = tag
	while k < j:
	    match = attrfind.match(rawdata, k)
	    if not match: break
	    attrname, rest, attrvalue = match.group(1, 2, 3)
	    if not rest:
		attrvalue = attrname
	    elif attrvalue[:1] == '\'' == attrvalue[-1:] or \
		 attrvalue[:1] == '"' == attrvalue[-1:]:
		attrvalue = attrvalue[1:-1]
	    attrs.append((string.lower(attrname), attrvalue))
	    k = match.end(0)
	if rawdata[j] == '>':
	    j = j+1
	self.finish_starttag(tag, attrs)
	return j

    # Internal -- parse endtag
    def parse_endtag(self, i):
	rawdata = self.rawdata
	match = endbracket.search(rawdata, i+1)
	if not match:
	    return -1
	j = match.start(0)
	tag = string.lower(string.strip(rawdata[i+2:j]))
	if rawdata[j] == '>':
	    j = j+1
	self.finish_endtag(tag)
	return j

    # Internal -- finish parsing of <tag/data/ (same as <tag>data</tag>)
    def finish_shorttag(self, tag, data):
	self.finish_starttag(tag, [])
	self.handle_data(data)
	self.finish_endtag(tag)

    # Internal -- finish processing of start tag
    # Return -1 for unknown tag, 0 for open-only tag, 1 for balanced tag
    def finish_starttag(self, tag, attrs):
	try:
	    method = getattr(self, 'start_' + tag)
	except AttributeError:
	    try:
		method = getattr(self, 'do_' + tag)
	    except AttributeError:
		self.unknown_starttag(tag, attrs)
		return -1
	    else:
		self.handle_starttag(tag, method, attrs)
		return 0
	else:
	    self.stack.append(tag)
	    self.handle_starttag(tag, method, attrs)
	    return 1

    # Internal -- finish processing of end tag
    def finish_endtag(self, tag):
	if not tag:
	    found = len(self.stack) - 1
	    if found < 0:
		self.unknown_endtag(tag)
		return
	else:
	    if tag not in self.stack:
		try:
		    method = getattr(self, 'end_' + tag)
		except AttributeError:
		    self.unknown_endtag(tag)
		return
	    found = len(self.stack)
	    for i in range(found):
		if self.stack[i] == tag: found = i
	while len(self.stack) > found:
	    tag = self.stack[-1]
	    try:
		method = getattr(self, 'end_' + tag)
	    except AttributeError:
		method = None
	    if method:
		self.handle_endtag(tag, method)
	    else:
		self.unknown_endtag(tag)
	    del self.stack[-1]

    # Overridable -- handle start tag
    def handle_starttag(self, tag, method, attrs):
	method(attrs)

    # Overridable -- handle end tag
    def handle_endtag(self, tag, method):
	method()

    # Example -- report an unbalanced </...> tag.
    def report_unbalanced(self, tag):
	if self.verbose:
	    print '*** Unbalanced </' + tag + '>'
	    print '*** Stack:', self.stack

    # Example -- handle character reference, no need to override
    def handle_charref(self, name):
	try:
	    n = string.atoi(name)
	except string.atoi_error:
	    self.unknown_charref(name)
	    return
	if not 0 <= n <= 255:
	    self.unknown_charref(name)
	    return
	self.handle_data(chr(n))

    # Definition of entities -- derived classes may override
    entitydefs = ENTITYDEFS

    # Example -- handle entity reference, no need to override
    def handle_entityref(self, name):
	table = self.entitydefs
	if table.has_key(name):
	    self.handle_data(table[name])
	else:
	    self.unknown_entityref(name)
	    return

    # Example -- handle data, should be overridden
    def handle_data(self, data):
	pass

    # Example -- handle comment, could be overridden
    def handle_comment(self, data):
	pass

    # To be overridden -- handlers for unknown objects
    def unknown_starttag(self, tag, attrs): pass
    def unknown_endtag(self, tag): pass
    def unknown_charref(self, ref): pass
    def unknown_entityref(self, ref): pass


# --------------------------------------------------------------------
# accelerated SGML parser

class FastSGMLParser:

    # Interface -- initialize and reset this instance
    def __init__(self, verbose=0):
	self.verbose = verbose
	self.reset()

    # Interface -- reset this instance.  Loses all unprocessed data
    def reset(self):
	self.rawdata = ''
	self.stack = []
	self.lasttag = '???'
	self.nomoretags = 0
	self.literal = 0
	self.parser = sgmlop.SGMLParser()
	self.feed = self.parser.feed
	self.parser.register(self)

    # For derived classes only -- enter literal mode (CDATA) till EOF
    def setnomoretags(self):
	self.nomoretags = self.literal = 1 # FIXME!

    # For derived classes only -- enter literal mode (CDATA)
    def setliteral(self, *args):
	self.literal = 1 # FIXME!

    # Interface -- feed some data to the parser.  Call this as
    # often as you want, with as little or as much text as you
    # want (may include '\n').
    def feed(self, data): # overridden by reset
	self.parser.feed(data)

    # Interface -- handle the remaining data
    def close(self):
	try:
	    self.parser.close()
	finally:
	    self.parser = None

    # Internal -- finish parsing of <tag/data/ (same as <tag>data</tag>)
    def finish_shorttag(self, tag, data):
	self.finish_starttag(tag, [])
	self.handle_data(data)
	self.finish_endtag(tag)

    # Internal -- finish processing of start tag
    # Return -1 for unknown tag, 0 for open-only tag, 1 for balanced tag
    def finish_starttag(self, tag, data):
	# FIXME: should use the sgmlop attribute parser!!!
	attrs = []
	k, j = 0, len(data)
	while k < j:
	    match = attrfind.match(data, k)
	    if not match: break
	    attrname, rest, attrvalue = match.group(1, 2, 3)
	    if not rest:
		attrvalue = attrname
	    elif attrvalue[:1] == '\'' == attrvalue[-1:] or \
		 attrvalue[:1] == '"' == attrvalue[-1:]:
		attrvalue = attrvalue[1:-1]
	    attrs.append((string.lower(attrname), attrvalue))
	    k = match.end(0)
	# FIXME: should move this logic into sgmlop!
	try:
	    method = getattr(self, 'start_' + tag)
	except AttributeError:
	    try:
		method = getattr(self, 'do_' + tag)
	    except AttributeError:
		self.unknown_starttag(tag, attrs)
		return -1
	    else:
		self.handle_starttag(tag, method, attrs)
		return 0
	else:
	    self.stack.append(tag)
	    self.handle_starttag(tag, method, attrs)
	    return 1

    # Internal -- finish processing of end tag
    def finish_endtag(self, tag):
	if not tag:
	    found = len(self.stack) - 1
	    if found < 0:
		self.unknown_endtag(tag)
		return
	else:
	    if tag not in self.stack:
		try:
		    method = getattr(self, 'end_' + tag)
		except AttributeError:
		    self.unknown_endtag(tag)
		return
	    found = len(self.stack)
	    for i in range(found):
		if self.stack[i] == tag: found = i
	while len(self.stack) > found:
	    tag = self.stack[-1]
	    try:
		method = getattr(self, 'end_' + tag)
	    except AttributeError:
		method = None
	    if method:
		self.handle_endtag(tag, method)
	    else:
		self.unknown_endtag(tag)
	    del self.stack[-1]

    # Overridable -- handle start tag
    def handle_starttag(self, tag, method, attrs):
	method(attrs)

    # Overridable -- handle end tag
    def handle_endtag(self, tag, method):
	method()

    # Example -- report an unbalanced </...> tag.
    def report_unbalanced(self, tag):
	if self.verbose:
	    print '*** Unbalanced </' + tag + '>'
	    print '*** Stack:', self.stack

    # Example -- handle character reference, no need to override
    # def handle_charref(self, name):
    #     pass

    # Definition of entities -- derived classes may override
    entitydefs = ENTITYDEFS

    # Example -- handle entity reference, no need to override
    def handle_entityref(self, name):
	try:
	    data = self.entitydefs[name]
	except KeyError:
	    self.unknown_entityref(name)
	else:
	    self.handle_data(data)

    # Example -- handle data, should be overridden
    def handle_data(self, data):
    	pass

    # Example -- handle comment, could be overridden
    # def handle_comment(self, data):
    #	pass

    # To be overridden -- handlers for unknown objects
    def unknown_starttag(self, tag, attrs): pass
    def unknown_endtag(self, tag): pass
    def unknown_charref(self, ref): pass
    def unknown_entityref(self, ref): pass


#sgmlop = None

# pick a suitable parser
if sgmlop:
    SGMLParser = FastSGMLParser
else:
    SGMLParser = SlowSGMLParser

# --------------------------------------------------------------------
# test stuff

class TestSGMLParser(SGMLParser):

    def __init__(self, verbose=0):
	self.testdata = ""
	SGMLParser.__init__(self, verbose)

    def handle_data(self, data):
	self.testdata = self.testdata + data
	if len(`self.testdata`) >= 70:
	    self.flush()

    def flush(self):
	data = self.testdata
	if data:
	    self.testdata = ""
	    print 'data:', `data`

    def handle_comment(self, data):
	self.flush()
	r = `data`
	if len(r) > 68:
	    r = r[:32] + '...' + r[-32:]
	print 'comment:', r

    def unknown_starttag(self, tag, attrs):
	self.flush()
	if not attrs:
	    print 'start tag: <' + tag + '>'
	else:
	    print 'start tag: <' + tag,
	    for name, value in attrs:
		print name + '=' + '"' + value + '"',
	    print '>'

    def unknown_endtag(self, tag):
	self.flush()
	print 'end tag: </' + tag + '>'

    def unknown_entityref(self, ref):
	self.flush()
	print '*** unknown entity ref: &' + ref + ';'

    def unknown_charref(self, ref):
	self.flush()
	print '*** unknown char ref: &#' + ref + ';'

    def close(self):
	SGMLParser.close(self)
	self.flush()


def test(args = None):
    import sys

    if not args:
	args = sys.argv[1:]

    if args and args[0] == '-s':
	args = args[1:]
	klass = SGMLParser
    else:
	klass = TestSGMLParser

    if args:
	file = args[0]
    else:
	file = 'test.html'

    if file == '-':
	f = sys.stdin
    else:
	try:
	    f = open(file, 'r')
	except IOError, msg:
	    print file, ":", msg
	    sys.exit(1)

    data = f.read()
    if f is not sys.stdin:
	f.close()

    x = klass()
    for c in data:
	x.feed(c)
    x.close()


if __name__ == '__main__':
    test()




from xml.dom.dc_parser import InferEndTags, Scanner
from xml.dom.core import DOMFactory
from xml.dom.builder import Builder


class DcBuilder(InferEndTags, Builder):
	
	def __init__(self):
		Builder.__init__(self)
		InferEndTags.__init__(self)
		self.scanner(Scanner())

	
	def feed(self, s):
		self.p.feed(s)
		self.p.next(self)
		self.eof()


	def startElement(self, elm):
		Builder.startElement(self, elm.name, elm.attrs)

	def endElement(self, elm):
		Builder.endElement(self, elm.name)


	def text(self, s):
		InferEndTags.text(self, s)
		Builder.text(self, s)




# Very simple test - Parse a file and print what happens
import sys
from xml.parsers import pyexpat
import sys
import os

class Outputter:
	def __init__(self, verbose=0):
		self.startcount = 0
		self.endcount = 0
		self.cdatacount = 0
		self.instcount = 0
		self.verbose = verbose

	def startelt(self, name, attrs):
		self.startcount = self.startcount + 1
		if self.verbose:
			print 'start', name
			for i in range(0, len(attrs), 2):
				print 'attr', attrs[i], attrs[i+1]

	def endelt(self, name):
		self.endcount = self.endcount + 1
		if self.verbose:
			print 'end', name

	def cdata(self, data):
		self.cdatacount = self.cdatacount + 1
		if self.verbose:
			print 'cdata', data

	def inst(self, target, data):
		self.instcount = self.instcount + 1
		if self.verbose:
			print 'inst', target, data

sys.argv[1:] = ["quotes.xml"]

out = Outputter(verbose = 1)
if len(sys.argv) != 2:
	if os.name == 'mac':
		import macfs
		fss, ok = macfs.StandardGetFile()
		if not ok: sys.exit(0)
		sys.argv.append(fss.as_pathname())
	else:
		print 'Usage: pyexpattest [-v] inputfile'
		sys.exit(1)
	
parser = pyexpat.ParserCreate()
parser.StartElementHandler = out.startelt
parser.EndElementHandler = out.endelt
parser.CharacterDataHandler = out.cdata
parser.ProcessingInstructionHandler = out.inst

data = open(sys.argv[1]).read()

rv = parser.Parse(data, 1)

print 'Parser returned', rv
if rv == 0:
	print '** Error', parser.ErrorCode
	print '** Line', parser.ErrorLineNumber
	print '** Column', parser.ErrorColumnNumber
	print '** Byte', parser.ErrorByteIndex
print 'Summary of XML parser upcalls:'
print 'start elements:', out.startcount
print 'end elements:', out.endcount
print 'character data:', out.cdatacount
print 'processing instructions:', out.instcount



from xml.dom.core import *

def explode(nodes, condition):
	for node in nodes:
		explode_node(node, condition)

	return nodes

def explode_node(node, condition):
	if node.NodeType != ELEMENT:
		return node

	explode(node.getChildren(), condition)

	if eval(condition, {}, {'this': node}):
		parent = node.getParentNode()
		for child in node.getChildren():
			parent.insertBefore(child, node)
		parent.removeChild(node)

	return node
		
		
if __name__ == '__main__':
	from xml.dom.pyhtml import *
	from xml.dom.writer import HtmlWriter

	tree = HTML(BODY(H1(A({'href':'blah'}, 'blah blah'))))
	w = HtmlWriter()
	w.write(tree)

	explode_node(tree, 'this.GI == "A"')
	w.write(tree)



			
			




'''transform.py: Document / grove transformation framework.

This module provides:

- a Transformer abstract base class, meant to be subclassed with ``do_*'' 
methods, where ``*'' stands for possible element tag names.

- various helper functions.

Issues: currently, the document or grove passed to a transformer is
tranformed in place, it might be better to keep the original.

'''

from xml.dom.core import *

def having(list, condition):
	'Filter nodes in <list> having <condition>.'
	#code = compile(condition)
	res = []
	for x in list:
		try:
			if eval(condition, {'this' : x}, {}):
				res.append(x)
		except:
			pass
	return res


def cdata(list):
	'Flatten a node list.'

	s = ''
	for node in list:
		if node.NodeType == TEXT:
			s = s + node.data
		elif node.NodeType == ELEMENT:
			s = s + cdata(node.getChildren())
	return s


class Transformer:
	simple_map = {}

	def __init__(self, factory=None):
		if factory:
			self.dom_factory = factory
		else:
			self.dom_factory = DOMFactory()


	def transform(self, x):
		'Tranform document or subtree.'

		if x.NodeType == DOCUMENT:
			root_element = x.documentElement
		else:
			root_element = x

		res = self._transform_node(root_element)[0]

		if x.NodeType == DOCUMENT:
			x.documentElement = res
			return x
		else:
			return res


	def _transform_node(self, node):
		'Transform node (currently, only elements are affected).'

		if node.NodeType == ELEMENT:
			new_children = []
			for child in node.getChildren():
				new_children = new_children + self._transform_node(child)
			node._children = new_children

			repl = self.simple_map.get(node.tagName)
			if repl:
				return self.transform_simple(node, repl)

			if hasattr(self, 'do_' + node.tagName):
				return getattr(self, 'do_' + node.tagName)(node)

			if hasattr(self, 'do_ANY'):
				return self.do_ANY(node)

		return [node]


	def transform_simple(self, node, repl):
		"Transform element using `simple' replacement."

		l = string.split(repl, '/')
		l.reverse()
		new_node = None
		for tag in l:
			new_new_node = self.dom_factory.createElement(tag, {})
			if new_node:
				new_node.insertBefore(new_new_node, None)
			else:
				for child in node.getChildren():
					new_new_node.insertBefore(child, None)
			new_node = new_new_node
		return [new_node]

# vim:ts=2:ai



"""DOM elements construction using Python syntax.
"""

from xml.dom import core

class Element(core.Element):
	tagName = GI = '???'
	is_empty = 0

	def __init__(self, *t, **d):
		core.Element.__init__(self, tagName=self.tagName)

		for name, value in d.items():
			self.set_attr(name, value)
			
		for x in t:
			if type(x) == type({}):
				for name, value in x.items():
					self.set_attr(name, value)
			elif type(x) in (type(()), type([])):
				for y in x:
					self.appendChild(y)
			elif type(x) == type(''):
				child = core.Text()
				child.data = x
				self.appendChild(child)
			else:
				self.appendChild(x)

	def __lshift__(self, other):
		self.insertBefore(other, None)


def makeElementClass(tagName):
	d = {}
	exec(
		'class %s(Element): tagName = GI = "%s"' % (tagName, tagName), 
		globals(), d
	)
	return d[tagName]


if __name__ == '__main__':
	HTML = makeElementClass('HTML')
	HEAD = makeElementClass('HEAD')
	BODY = makeElementClass('BODY')
	H1 = makeElementClass('H1')

	t = HTML(HEAD(), BODY(H1('this is a test'), "don't panic."))
	from xml.dom.writer import HtmlWriter
	l = HtmlWriter()
	l.write(t)
